/*
 * Copyright (C) 2023, IOPSYS Software Solutions AB.
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdbool.h>

#include "os_utils.h"
#include "text_utils.h"

#include "vendor_ubus_thread.h"
#include "vendor_utils.h"

struct ubus_thread_global {
       int tid;
       int mq_sockets[2];
       struct ubus_context *ubus_ctx; // ubus context
       struct ubus_event_handler ubus_notify_event; // ubus usp.event to notify about events
       struct ubus_event_handler ubus_add_event;    // ubus event to update add_instances
       struct ubus_event_handler ubus_del_event;    // ubus event to delete instance
       struct uloop_timeout instance_refresh_timer;
       str_vector_t fault_instances;
       pthread_cond_t ubus_cond;
       pthread_mutex_t ubus_lock;
};

static struct ubus_thread_global g_ubus_data;

#define MQ_RX_SOCKET  g_ubus_data.mq_sockets[0]
#define MQ_TX_SOCKET  g_ubus_data.mq_sockets[1]
#define STR(s) #s

#define DM_EVENT DM_OBJECT_NAME ".event"
#define DM_ADD_OBJ_EV_NAME DM_OBJECT_NAME ".AddObj"
#define DM_DEL_OBJ_EV_NAME DM_OBJECT_NAME ".DelObj"
#define MAX_TRANSACTION_RETRY (5)
#define USP_PROTO "usp"
#define MIN_NUM_TO_GROUP (10)

static void _pull_root_instances(str_vector_t *inst_vec);
int _uspd_call(struct ubus_context *ctx, char *object, char *method,
		struct blob_buf *data, ubus_data_handler_t callback,
		vendor_data_t *arg);

static int ubus_data_init()
{
	memset(&g_ubus_data, 0, sizeof(struct ubus_thread_global));

	g_ubus_data.ubus_ctx = ubus_connect(NULL);
	if (g_ubus_data.ubus_ctx == NULL) {
		USP_LOG_Error("[%s:%d] ubus_connect failed %d", __func__, __LINE__, errno);
		return USP_ERR_INTERNAL_ERROR;
	}

	// Required in init
	STR_VECTOR_Init(&g_ubus_data.fault_instances);

	g_ubus_data.mq_sockets[0] = -1;
	g_ubus_data.mq_sockets[1] = -1;
	g_ubus_data.tid = INVALID;

	pthread_cond_init(&g_ubus_data.ubus_cond, NULL);
	pthread_mutex_init(&g_ubus_data.ubus_lock, NULL);

	return USP_ERR_OK;
}

static int ubus_data_free()
{
	if (g_ubus_data.ubus_ctx) {
		ubus_free(g_ubus_data.ubus_ctx);
		g_ubus_data.ubus_ctx = NULL;
	}

	STR_VECTOR_Destroy(&g_ubus_data.fault_instances);

	pthread_cond_destroy(&g_ubus_data.ubus_cond);
	pthread_mutex_destroy(&g_ubus_data.ubus_lock);

	return USP_ERR_OK;
}

// Event handling
static void event_handler(struct ubus_context *ctx, struct ubus_event_handler *ev, const char *type, struct blob_attr *msg)
{
	char *ename;
	kv_vector_t *eparam;
	dm_node_t *node;
	enum {
		P_NAME,
		P_INPUT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "name", BLOBMSG_TYPE_STRING },
		{ "input", BLOBMSG_TYPE_TABLE },
	};
	struct blob_attr *tb[__P_MAX] = {NULL};

	if (blobmsg_parse(p, __P_MAX, tb, blobmsg_data(msg), blobmsg_len(msg)) != UBUS_STATUS_OK) {
		USP_LOG_Info("Failed to parse ubus data");
		return;
	}

	if (!tb[P_NAME]) {
		USP_LOG_Info("Event name not present");
	}
	ename = blobmsg_get_string(tb[P_NAME]);
	node = DM_PRIV_GetNodeFromPath(ename, NULL, NULL);
	if (node == NULL) {
		USP_LOG_Info("Event/Param (%s) not registered", ename);
		return;
	}

	if (node->type != kDMNodeType_Event) {
		USP_LOG_Info("Parameter (%s) not an event", ename);
		return;
	}

	eparam = blob_table_to_kv(tb[P_INPUT]);
	// Dump parameters for cli validation
	if (usp_log_level >= kLogLevel_Info) {
		KV_VECTOR_Dump(eparam);
	}
	USP_SIGNAL_DataModelEvent(ename, eparam);
}

static void event_handler_add_del(struct ubus_context *ctx, struct ubus_event_handler *ev, const char *type, struct blob_attr *msg)
{
	bool is_add = false;
	struct blob_attr *cur;
	int rem;
	enum {
		P_PARAMS,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "instances", BLOBMSG_TYPE_ARRAY },
	};
	struct blob_attr *tb[__P_MAX] = {NULL};

	if (blobmsg_parse(p, __P_MAX, tb, blobmsg_data(msg), blobmsg_len(msg)) != UBUS_STATUS_OK) {
		USP_LOG_Info("Failed to parse ubus data");
		return;
	}

	if (!tb[P_PARAMS]) {
		USP_LOG_Info("Event %s has no instances to update", type);
		return;
	}

	if (strcmp(type, DM_ADD_OBJ_EV_NAME) == 0)
		is_add = true;

	blobmsg_for_each_attr(cur, tb[P_PARAMS], rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING)
			continue;

		USP_LOG_Debug("Event %s :: %s", (is_add)?"ADD":"DEL", blobmsg_get_string(cur));
		if (is_add)
			USP_SIGNAL_ObjectAdded(blobmsg_get_string(cur));
		else
			USP_SIGNAL_ObjectDeleted(blobmsg_get_string(cur));
	}
}

static void _handle_faulty_keys(char *key)
{
	char buffer[MAX_DM_PATH] = {0};
	char *cinst;
	size_t llen, slen;

	slen = strlen(key);
	TEXT_UTILS_PathToSchemaForm(key, buffer, sizeof(buffer));
	cinst = strrchr(buffer, '}');
	if (cinst == NULL) {
		return;
	}
	llen = strlen(buffer) - abs(cinst - buffer) - 1;
	USP_STRNCPY(buffer, key, slen - llen + 1);
	STR_VECTOR_Add_IfNotExist(&g_ubus_data.fault_instances, buffer);
}

static void _check_delete_instance(struct uloop_timeout *t)
{
	int i;
	str_vector_t inst_vec;

	STR_VECTOR_Init(&inst_vec);

	// Pull instances from backend
	_pull_root_instances(&inst_vec);
	for (i = 0; i < g_ubus_data.fault_instances.num_entries; i++) {
		int found = INVALID;

		found = STR_VECTOR_Find(&inst_vec, g_ubus_data.fault_instances.vector[i]);
		if (found == INVALID) {
			USP_LOG_Debug("# Deleting inst [%s]", g_ubus_data.fault_instances.vector[i]);
			USP_SIGNAL_ObjectDeleted(g_ubus_data.fault_instances.vector[i]);
		} else {
			USP_LOG_Error("## inst [%s] still exists in backend", g_ubus_data.fault_instances.vector[i]);
		}
	}
	STR_VECTOR_Destroy(&inst_vec);
	STR_VECTOR_Destroy(&g_ubus_data.fault_instances);
}

// Registration handling
static char *stringify_cmd(int cmd)
{
	switch(cmd) {
	case CMD_NULL:
		return STR(CMD_NULL);
	case CMD_GROUP_GET:
		return STR(CMD_GROUP_GET);
	case CMD_GROUP_SET:
		return STR(CMD_GROUP_SET);
	case CMD_ADD:
		return STR(CMD_ADD);
	case CMD_DEL:
		return STR(CMD_DEL);
	case CMD_GET:
		return STR(CMD_GET);
	case CMD_SET:
		return STR(CMD_SET);
	case CMD_OPERATE_SYNC:
		return STR(CMD_OPERATE_SYNC);
	case CMD_OPERATE_ASYNC:
		return STR(CMD_OPERATE_ASYNC);
	case CMD_INSTANCES:
		return STR(CMD_INSTANCES);
	case CMD_TRAN_START:
		return STR(CMD_TRAN_START);
	case CMD_TRAN_COMMIT:
		return STR(CMD_TRAN_COMMIT);
	case CMD_TRAN_ABORT:
		return STR(CMD_TRAN_ABORT);
	case CMD_STOP:
		return STR(CMD_STOP);
	}

	return NULL;
}

static void _get_value_single_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	// No place holder to return value for the variable, returning
	if (arg->p_value) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg, "results");
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		char path[MAX_DM_PATH], val[MAX_DM_PATH];
		int fault;

		fault = get_details_from_blob(cur, path, val, NULL, NULL);
		if (fault != USP_ERR_OK) {
			arg->fault = fault;
		}
		if (strcmp(path, arg->path) == 0) {
			USP_STRNCPY(arg->p_value, val, arg->len);
			break;
		}
	}
}

static int _get_value_single(vendor_data_t *arg)
{
	struct blob_buf bb = { };

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blob_add_optinal_params(&bb, INVALID);

	// Invoke Ubus to get data from uspd
	uspd_call(g_ubus_data.ubus_ctx, "get", &bb, _get_value_single_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static void _get_value_group_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	int index, i;
	vendor_data_t *arg = req->priv;
	struct blob_attr *params, *cur;
	size_t rem;
	kv_vector_t pv_set;

	USP_ARG_Init(&pv_set);
	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg, "results");
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	index = 0;
	blobmsg_for_each_attr(cur, params, rem) {
		char path[MAX_DM_PATH], val[MAX_DM_PATH];
		int fault;

		fault = get_details_from_blob(cur, path, val, NULL, NULL);
		if (fault != USP_ERR_OK) {
			arg->fault = fault;
		}
		USP_ARG_Add(&pv_set, path, val);
	}

	for (i = 0, index = 0; i < arg->kv->num_entries; i++) {
		kv_pair_t *pair = &arg->kv->vector[i];

		index = KV_VECTOR_FindKey(&pv_set, pair->key, index);
		if (index == INVALID) {
			_handle_faulty_keys(pair->key);
		} else {
			// Found a matching key, so replace its value
			USP_SAFE_FREE(pair->value);
			pair->value = USP_STRDUP(pv_set.vector[index].value);
			index++;
		}
	}

	if (arg->kv_out) {
		arg->kv_out->num_entries = pv_set.num_entries;
		arg->kv_out->vector = pv_set.vector;
		USP_ARG_Init(&pv_set);
	}

	USP_ARG_Destroy(&pv_set);
}

static int get_min_number_to_group(void)
{
	int ret = MIN_NUM_TO_GROUP;
	char value[25] = {0};
	char uci_path[] = "obuspa.global.min_num_to_group";

	get_value_from_uci(uci_path, value, sizeof(value));
	if (value[0] != '\0') {
		int val = strtod(value, NULL);
		if (val > 0)
			ret = val;
	}

	return ret;
}

static int _get_value_group(vendor_data_t *arg)
{
	struct blob_buf bb = { };
	char *method, *p;
	void *array;
	int i;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	if (g_ubus_data.fault_instances.num_entries) {
		STR_VECTOR_Destroy(&g_ubus_data.fault_instances);
	}

	method = "get";
	blob_add_optinal_params(&bb, INVALID);
	int min_num_to_group = get_min_number_to_group();
	if (strcmp(arg->path, ROOT_PATH) == 0) {
		blobmsg_add_string(&bb, "path", arg->path);
	} else if (arg->kv->num_entries < min_num_to_group) {
		array = blobmsg_open_array(&bb, "paths");
		for (i = 0; i < arg->kv->num_entries; ++i) {
			p = arg->kv->vector[i].key;
			blobmsg_add_string(&bb, NULL, p);
		}
		blobmsg_close_array(&bb, array);
	} else {
		blobmsg_add_string(&bb, "path", arg->path);
	}

	uspd_call(g_ubus_data.ubus_ctx, method, &bb, _get_value_group_cb, arg);

	if (g_ubus_data.fault_instances.num_entries) {
		g_ubus_data.instance_refresh_timer.cb = _check_delete_instance;
		uloop_timeout_set(&g_ubus_data.instance_refresh_timer, 0);
	}

	blob_buf_free(&bb);
	return arg->fault;
}

static void _get_instances_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *cur, *params = NULL;
	size_t rem;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg, "results");
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		char path[MAX_DM_PATH];
		int fault;

		fault = get_details_from_blob(cur, path, NULL, NULL, NULL);
		if (fault != USP_ERR_OK) {
			arg->fault = fault;
		}

		STR_VECTOR_Add(arg->vec, path);
	}
}

static int _get_instances(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blob_add_optinal_params(&bb, INVALID);

	uspd_call(g_ubus_data.ubus_ctx, DM_INSTANCE_METHOD, &bb, _get_instances_cb, arg);

	blob_buf_free(&bb);

	return USP_ERR_OK;
}

static void _pull_root_instances(str_vector_t *inst_vec)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_INSTANCES;
	arg.path = ROOT_PATH;
	arg.vec = inst_vec;

	_get_instances(&arg);
}

static void _add_object_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	size_t rem;
	struct blob_attr *results, *cur;
	vendor_data_t *arg = req->priv;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	results = get_parameters(msg, "results");
	if (results == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, results, rem) {
		char instance[MAX_DM_PATH] = {0};

		arg->fault = get_details_from_blob(cur, NULL, instance, NULL, NULL);
		*arg->p_instance = atoi(instance);
		break;
	}

}

static int _add_object(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blob_add_optinal_params(&bb, g_ubus_data.tid);

	uspd_call(g_ubus_data.ubus_ctx, "add", &bb, _add_object_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static void _set_value_single_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	size_t rem;
	struct blob_attr *results, *cur;
	vendor_data_t *arg = req->priv;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	results = get_parameters(msg, "results");
	if (results == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, results, rem) {
		arg->fault = get_details_from_blob(cur, NULL, NULL, NULL, NULL);
		break;
	}
}

static int _set_value_single(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "value", arg->p_value);
	blob_add_optinal_params(&bb, g_ubus_data.tid);

	uspd_call(g_ubus_data.ubus_ctx, "set", &bb, _set_value_single_cb, arg);

	blob_buf_free(&bb);
	return USP_ERR_OK;
}

static void _set_value_group_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	vendor_data_t *arg = req->priv;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg, "results");
	if (params == NULL) {
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		char path[MAX_DM_PATH];

		arg->fault = get_details_from_blob(cur, path, NULL, NULL, NULL);
		if ( arg->fault != USP_ERR_OK) {
			*arg->p_failure_index = KV_VECTOR_FindKey(arg->kv, path, 0);
		}
	}
}

static int _set_value_group(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	void *table;
	int i;
	size_t path_len;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	path_len = strlen(arg->path);
	blobmsg_add_string(&bb, "path", arg->path);
	table = blobmsg_open_table(&bb, "obj_path");
	for (i = 0; i < arg->kv->num_entries; i++) {
		blobmsg_add_string(&bb, arg->kv->vector[i].key + path_len, arg->kv->vector[i].value);
	}
	blobmsg_close_table(&bb, table);

	blob_add_optinal_params(&bb, g_ubus_data.tid);

	uspd_call(g_ubus_data.ubus_ctx, "set", &bb, _set_value_group_cb, arg);

	blob_buf_free(&bb);
	return USP_ERR_OK;
}

static void _del_object_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	vendor_data_t *arg = req->priv;

	if (!msg) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	params = get_parameters(msg, "results");
	if (params == NULL) {
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_for_each_attr(cur, params, rem) {
		arg->fault = get_details_from_blob(cur, NULL, NULL, NULL, NULL);
	}
}

static int _del_object(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blob_add_optinal_params(&bb, g_ubus_data.tid);

	uspd_call(g_ubus_data.ubus_ctx, "del", &bb, _del_object_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static void _sync_operate_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *results, *rtemp;
	kv_vector_t *kv_out;
	size_t rem;

	if (!msg)
		return;

	kv_out = arg->kv_out;

	results = get_parameters(msg, "results");
	if (results == NULL)
		return;

	blobmsg_for_each_attr(rtemp, results, rem) {
		arg->fault = get_details_from_blob(rtemp, NULL, NULL, NULL, NULL);
		if (arg->fault != USP_ERR_OK)
			return;

		struct blob_attr *params = get_parameters(rtemp, "output");

		if (params == NULL)
			return;

		get_parameters_in_kv(params, kv_out);
	}
}

static int _operate_sync_handler(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	if (arg->kv_out)
		KV_VECTOR_Init(arg->kv_out);

	blobmsg_add_string(&bb, "command", arg->path);
	if (arg->p_value) {
		blobmsg_add_string(&bb, "command_key", arg->p_value);
	}

	if (arg->kv != NULL) {
		void *table = blobmsg_open_table(&bb, "input");

		for (int i = 0; i < arg->kv->num_entries; ++i) {
			kv_pair_t *kv;

			kv = &arg->kv->vector[i];
			blobmsg_add_string(&bb, kv->key, kv->value);
		}
		blobmsg_close_table(&bb, table);
	}

	blob_add_optinal_params(&bb, g_ubus_data.tid);

	uspd_call(g_ubus_data.ubus_ctx, "operate", &bb, _sync_operate_cb, arg);

	blob_buf_free(&bb);

	return arg->fault;
}

static void _operate_async_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *results, *rtemp;
	int *instance = (int *) req->priv;
	char err_log[128] = {0};
	char *err_msg;
	int fault = USP_ERR_OK;
	kv_vector_t *output_args;
	size_t rem;

	output_args = USP_ARG_Create();
	if (!msg) {
		fault = USP_ERR_INTERNAL_ERROR;
		goto exit;
	}

	results = get_parameters(msg, "results");
	if (results == NULL) {
		fault = USP_ERR_INTERNAL_ERROR;
		goto exit;
	}

	blobmsg_for_each_attr(rtemp, results, rem) {
		fault = get_details_from_blob(rtemp, NULL, NULL, NULL, err_log);
		if (fault != USP_ERR_OK)
			break;

		struct blob_attr *params = get_parameters(rtemp, "output");

		if (params == NULL)
			break;

		get_parameters_in_kv(params, output_args);
	}

exit:
	if (usp_log_level >= kLogLevel_Info) {
		USP_LOG_Info("## Dump async out arg:");
		KV_VECTOR_Dump(output_args);
	}

	err_msg = (fault != USP_ERR_OK) ? err_log : NULL;
	USP_SIGNAL_OperationComplete(*instance, fault, err_msg, output_args);
	USP_SAFE_FREE(instance);
}

static int _operate_async_handler(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	int fault = USP_ERR_OK;
	int *instance;

	instance = (int *) USP_MALLOC(sizeof(int));
	if (instance == NULL)
		return USP_ERR_COMMAND_FAILURE;

	// Exit if unable to signal that this operation is active
	fault = USP_SIGNAL_OperationStatus(arg->inst, "Active");
	if (fault != USP_ERR_OK) {
		USP_FREE(instance);
		return USP_ERR_COMMAND_FAILURE;
	}

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "command", arg->path);
	blob_add_optinal_params(&bb, g_ubus_data.tid);

	if (arg->kv != NULL) {
		void *table = blobmsg_open_table(&bb, "input");

		for (int i = 0; i < arg->kv->num_entries; ++i) {
			kv_pair_t *kv = &arg->kv->vector[i];

			blobmsg_add_string(&bb, kv->key, kv->value);
		}
		blobmsg_close_table(&bb, table);
	}

	*instance = arg->inst;
	fault = uspd_call_async("operate", &bb, _operate_async_cb, instance);
	if (fault != USP_ERR_OK) {
		USP_FREE(instance);
		arg->fault = fault;
	}

	blob_buf_free(&bb);
	return fault;
}

static void _tran_exec_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	bool status = false;
	struct blob_attr *f_attr[1] = {NULL};
	const struct blobmsg_policy fp[1] = {
		{ "status", BLOBMSG_TYPE_BOOL }
	};

	g_ubus_data.tid = 0;

	if (!msg) {
		arg->fault = USP_ERR_OK;
		return;
	}

	blobmsg_parse(fp, 1, f_attr, blobmsg_data(msg), blobmsg_len(msg));
	if (f_attr[0])
		status = blobmsg_get_u8(f_attr[0]);

	if (status == false) {
		USP_LOG_Error("Another transaction in progress");
		arg->fault = USP_ERR_OK;
	}
}

static void _tran_start_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	bool status = false;
	int tid = 0;
	struct blob_attr *f_attr[2] = {NULL};
	const struct blobmsg_policy fp[2] = {
		{ "status", BLOBMSG_TYPE_BOOL },
		{ "transaction_id", BLOBMSG_TYPE_INT32 }
	};

	if (!msg) {
		g_ubus_data.tid = 0;
		arg->fault = USP_ERR_INTERNAL_ERROR;
		return;
	}

	blobmsg_parse(fp, 2, f_attr, blobmsg_data(msg), blobmsg_len(msg));

	if (f_attr[0])
		status = blobmsg_get_u8(f_attr[0]);

	if (f_attr[1])
		tid = blobmsg_get_u32(f_attr[1]);

	if (status == false) {
		g_ubus_data.tid = 0;
		USP_LOG_Info("Another transaction in progress");
		arg->fault = USP_ERR_INTERNAL_ERROR;
	} else {
		g_ubus_data.tid = tid;
	}
}

static int _tran_exec(vendor_data_t *arg, char *cmd)
{
	int retry = 0;
	struct blob_buf bb = { };

	// Ignore commit if its in start sequence
	if ((strcmp(cmd, "start") != 0) && g_ubus_data.tid == INVALID) {
		USP_LOG_Debug("Ignoring %s, as tid is %d", cmd, g_ubus_data.tid);
		return USP_ERR_OK;
	}

TSTART:
	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	// Invoke Ubus to get data from uspd
	blobmsg_add_string(&bb, "cmd", cmd);
	blobmsg_add_u32(&bb, "timeout", arg->ipc_timeout);
	blobmsg_add_u8(&bb, "restart_services", true);

	if (strcmp(cmd, "start") == 0) {
		blob_add_optinal_params(&bb, INVALID);
		uspd_call(g_ubus_data.ubus_ctx, "transaction", &bb, _tran_start_cb, arg);
	} else {
		blob_add_optinal_params(&bb, g_ubus_data.tid);
		uspd_call(g_ubus_data.ubus_ctx, "transaction", &bb, _tran_exec_cb, arg);
	}
	blob_buf_free(&bb);

	// In case of fault, retry till MAX_TRANSACTION_RETRY
	if (arg->fault != USP_ERR_OK && retry < MAX_TRANSACTION_RETRY) {
		sleep(2);
		retry++;
		goto TSTART;
	}

	return USP_ERR_OK;
}

static int stop_service()
{
	struct blob_buf bb = { };

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	// Invoke Ubus to get data from uspd
	blobmsg_add_string(&bb, "name", "obuspa");
	blobmsg_add_string(&bb, "instance", "obuspa");

	_uspd_call(g_ubus_data.ubus_ctx, "service", "delete", &bb, NULL, NULL);
	blob_buf_free(&bb);

	return USP_ERR_OK;
}

static void read_cmd_callback(struct uloop_fd *u, unsigned int events)
{
	vendor_data_t *arg = NULL;
	int bytes_read;

	pthread_mutex_lock(&g_ubus_data.ubus_lock);
	bytes_read = recv(MQ_RX_SOCKET, &arg, sizeof(vendor_data_t *), 0);
	if (bytes_read != sizeof(vendor_data_t *)) {
		USP_LOG_Error("[UBUS] size mismatch %d", bytes_read);
		pthread_cond_signal(&g_ubus_data.ubus_cond);
		pthread_mutex_unlock(&g_ubus_data.ubus_lock);
		return;
	}

	switch (arg->cmd) {
	case CMD_GET:
		_get_value_single(arg);
		break;
	case CMD_GROUP_GET:
		_get_value_group(arg);
		break;
	case CMD_INSTANCES:
		_get_instances(arg);
		break;
	case CMD_ADD:
		_add_object(arg);
		break;
	case CMD_SET:
		_set_value_single(arg);
		break;
	case CMD_GROUP_SET:
		_set_value_group(arg);
		break;
	case CMD_DEL:
		_del_object(arg);
		break;
	case CMD_OPERATE_SYNC:
		_operate_sync_handler(arg);
		break;
	case CMD_OPERATE_ASYNC:
		_operate_async_handler(arg);
		break;
	case CMD_TRAN_START:
		_tran_exec(arg, "start");
		break;
	case CMD_TRAN_COMMIT:
		_tran_exec(arg, "commit");
		break;
	case CMD_TRAN_ABORT:
		_tran_exec(arg, "abort");
		break;
	case CMD_STOP:
		stop_service();
		break;
	case CMD_NULL:
		USP_LOG_Info("Command not supported as of now");
		break;
	}

	pthread_cond_signal(&g_ubus_data.ubus_cond);
	pthread_mutex_unlock(&g_ubus_data.ubus_lock);
}


static void *start_ubus_thread(void *arg)
{
	struct uloop_fd cmd_fd;

	pthread_mutex_lock(&g_ubus_data.ubus_lock);
	uloop_init();
	ubus_add_uloop(g_ubus_data.ubus_ctx);

	memset(&cmd_fd, 0, sizeof(struct uloop_fd));
	cmd_fd.cb = read_cmd_callback;
	cmd_fd.fd = MQ_RX_SOCKET;
	uloop_fd_add(&cmd_fd, ULOOP_READ | ULOOP_EDGE_TRIGGER);

	// register for ubus events
	g_ubus_data.ubus_notify_event.cb = event_handler;
	ubus_register_event_handler(g_ubus_data.ubus_ctx, &g_ubus_data.ubus_notify_event, DM_EVENT);

	g_ubus_data.ubus_add_event.cb = event_handler_add_del;
	ubus_register_event_handler(g_ubus_data.ubus_ctx, &g_ubus_data.ubus_add_event, DM_ADD_OBJ_EV_NAME);
	g_ubus_data.ubus_del_event.cb = event_handler_add_del;
	ubus_register_event_handler(g_ubus_data.ubus_ctx, &g_ubus_data.ubus_del_event, DM_DEL_OBJ_EV_NAME);

	pthread_cond_signal(&g_ubus_data.ubus_cond);
	pthread_mutex_unlock(&g_ubus_data.ubus_lock);
	uloop_run();

	uloop_done();

	close(MQ_TX_SOCKET);
	close(MQ_RX_SOCKET);
	ubus_data_free();

	return NULL;
}

static int spawn_ubus_thread(void)
{
	int fault = USP_ERR_OK;

	pthread_mutex_lock(&g_ubus_data.ubus_lock);
	// Exit if unable to initialize the unix domain socket pair used to implement a message queue
	fault = socketpair(AF_UNIX, SOCK_DGRAM, 0, g_ubus_data.mq_sockets);
	if (fault != 0) {
		USP_LOG_Error("Failed to create socketpair...");
		return USP_ERR_INTERNAL_ERROR;
	}

	fault = OS_UTILS_CreateThread("UBUS_Thread", start_ubus_thread, NULL);
	if (fault != 0) {
		USP_LOG_Error("Failed to uloop thread ...");
		return USP_ERR_INTERNAL_ERROR;
	}
	pthread_cond_wait(&g_ubus_data.ubus_cond, &g_ubus_data.ubus_lock);
	pthread_mutex_unlock(&g_ubus_data.ubus_lock);

	return fault;
}

static void uspd_ubus_complete_cb(struct ubus_request *req, int ret)
{
    USP_FREE(req);
}

int _uspd_call(struct ubus_context *ctx, char *object, char *method,
		struct blob_buf *data, ubus_data_handler_t callback,
		vendor_data_t *arg)
{
	uint32_t id;
	int fault = UBUS_STATUS_OK;

	fault = ubus_lookup_id(ctx, object, &id);
	if (fault) {
		USP_LOG_Error("[%s:%d] %s not present (%s)",
			      __func__, __LINE__, object, ubus_strerror(fault));
		if (arg)
			arg->fault = USP_ERR_INTERNAL_ERROR;
		return USP_ERR_INTERNAL_ERROR;
	}

	// Invoke Ubus to get data from uspd
	fault = ubus_invoke(ctx, id, method, data->head, callback, arg, (arg)?arg->ipc_timeout:DEFAULT_IPC_TIMEOUT);
	if (fault) {
		USP_LOG_Error("[%s:%d] ubus call for (%s)method failed(%s)",
			      __func__, __LINE__, method, ubus_strerror(fault));
		if (arg)
			arg->fault = USP_ERR_INTERNAL_ERROR;
		return USP_ERR_INTERNAL_ERROR;
	}

	return USP_ERR_OK;
}

// Public methods
int uspd_call(struct ubus_context *ctx, char *method,
		struct blob_buf *data, ubus_data_handler_t callback,
		vendor_data_t *arg)
{
	return _uspd_call(ctx, DM_OBJECT_NAME, method, data, callback, arg);
}

int uspd_call_async(char *method, struct blob_buf *data, ubus_data_handler_t callback, void *priv)
{
	uint32_t id;
	int fault = UBUS_STATUS_OK;
	struct ubus_request *req;

	fault = ubus_lookup_id(g_ubus_data.ubus_ctx, DM_OBJECT_NAME, &id);
	if (fault) {
		USP_LOG_Error("[%s:%d] %s not present (%s)",
			      __func__, __LINE__, DM_OBJECT_NAME, ubus_strerror(fault));
		return USP_ERR_INTERNAL_ERROR;
	}

	req = USP_MALLOC(sizeof(struct ubus_request));
	if (req == NULL) {
		USP_LOG_Error("Out of memory!");
		return USP_ERR_RESOURCES_EXCEEDED;
	}

	memset(req, 0, sizeof(struct ubus_request));
	fault = ubus_invoke_async(g_ubus_data.ubus_ctx, id, method, data->head, req);
	if (fault) {
		USP_LOG_Error("[%s:%d] ubus async call failed(%s)",
			      __func__, __LINE__, ubus_strerror(fault));
		USP_FREE(req);
		return USP_ERR_INTERNAL_ERROR;
	}

	req->data_cb = callback;
	req->complete_cb = uspd_ubus_complete_cb;
	req->priv = priv;

	ubus_complete_request_async(g_ubus_data.ubus_ctx, req);

	return USP_ERR_OK;
}

int ubus_enque_cmd(vendor_data_t *arg)
{
	USP_LOG_Debug("## CMD (%s) path(%s) Socket TX:RX(%d:%d) ##", stringify_cmd(arg->cmd), arg->path, MQ_TX_SOCKET, MQ_RX_SOCKET);

	if (MQ_TX_SOCKET != -1) {
		int bytes;

		pthread_mutex_lock(&g_ubus_data.ubus_lock);
		bytes = send(MQ_TX_SOCKET, &arg, sizeof(vendor_data_t *), 0);
		if (bytes == -1) {
			pthread_mutex_unlock(&g_ubus_data.ubus_lock);
			USP_LOG_Error("## Failed to send err (%s)", strerror(errno));
			return USP_ERR_INTERNAL_ERROR;
		}

		if (bytes != sizeof(vendor_data_t *)) {
			pthread_mutex_unlock(&g_ubus_data.ubus_lock);
			USP_LOG_Error("[DMThread] send size miss-match (%d:%zd) to ubus thread", bytes, sizeof(vendor_data_t *));
			return USP_ERR_INTERNAL_ERROR;
		}
		pthread_cond_wait(&g_ubus_data.ubus_cond, &g_ubus_data.ubus_lock);
		pthread_mutex_unlock(&g_ubus_data.ubus_lock);

		return arg->fault;
	}

	switch (arg->cmd) {
	case CMD_GET:
		_get_value_single(arg);
		break;
	case CMD_GROUP_GET:
		_get_value_group(arg);
		break;
	case CMD_TRAN_START:
		USP_LOG_Debug("Ignoring transaction start");
		break;
	case CMD_TRAN_COMMIT:
		USP_LOG_Debug("Ignoring transaction commit");
		break;
	case CMD_TRAN_ABORT:
		USP_LOG_Debug("Ignoring transaction abort");
		break;
	default:
		USP_LOG_Error("%s(%d): uspd thread not avilable", __FUNCTION__, __LINE__);
		arg->fault = USP_ERR_INTERNAL_ERROR;
		break;
	}

	return arg->fault;
}

int ubus_thread_init()
{
	return ubus_data_init();
}

int ubus_thread_start(void)
{
	int fault = USP_ERR_OK;

	fault = spawn_ubus_thread();

	return fault;
}
